# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

if __name__ == "__main__":
    import unittest, os

    suffix = "__main__.py"
    assert __file__.endswith(suffix)
    test_dir = __file__[:-len(suffix)]

    skip_l = int(os.environ.get("test_skip_l", "0"))
    skip_r = int(os.environ.get("test_skip_r", "1000000"))
    test_only = None
    if "test_only" in os.environ:
        test_only = set(os.environ.get("test_only").split(","))

    test_files = os.listdir(test_dir)
    test_files = sorted(test_files)
    suite = unittest.TestSuite()
    
    for _, test_file in enumerate(test_files):
        if not test_file.startswith("test_"):
            continue
        if _ < skip_l or _ > skip_r:
            continue
        test_name = test_file.split(".")[0]
        if test_only and test_name not in test_only:
            continue

        print("Add Test", _, test_name)
        suite.addTest(unittest.defaultTestLoader.loadTestsFromName(
            "jittor.test."+test_name))

    unittest.TextTestRunner(verbosity=3).run(suite)


import jittor as jt
from jittor import nn
from jittor import Module
from jittor import init
from jittor.contrib import concat, argmax_pool
import math

class Bottle2neck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None, baseWidth=26, scale = 4, stype='normal'):
        """ Constructor
        Args:
            inplanes: input channel dimensionality
            planes: output channel dimensionality
            stride: conv stride. Replaces pooling layer.
            downsample: None when stride = 1
            baseWidth: basic width of conv3x3
            scale: number of scale.
            type: 'normal': normal set. 'stage': first block of a new stage.
        """
        super(Bottle2neck, self).__init__()

        width = int(math.floor(planes * (baseWidth/64.0)))
        self.conv1 = nn.Conv(inplanes, width*scale, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm(width*scale)
        assert scale > 1, 'Res2Net degenerates to ResNet when scales = 1.'
        if scale == 1:
          self.nums = 1
        else:
          self.nums = scale -1
        if stype == 'stage':
            self.pool = nn.Pool(kernel_size=3, stride = stride, padding=1, op='mean')
        self.convs = nn.ModuleList()
        self.bns = nn.ModuleList()
        for i in range(self.nums):
          self.convs.append(nn.Conv(width, width, kernel_size=3, stride = stride, dilation=dilation, padding=dilation, bias=False))
          self.bns.append(nn.BatchNorm(width))

        self.conv3 = nn.Conv(width*scale, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm(planes * self.expansion)

        self.relu = nn.ReLU()
        self.downsample = downsample
        self.stype = stype
        self.scale = scale
        self.width  = width
        self.stride = stride
        self.dilation = dilation

    def execute(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        spx = out
        
        outs = []
        for i in range(self.nums):
            if i==0 or self.stype=='stage':
                sp = spx[:, i*self.width: (i+1)*self.width]
            else:
                sp = sp + spx[:, i*self.width: (i+1)*self.width]
            sp = self.convs[i](sp)
            sp = self.relu(self.bns[i](sp))
            outs.append(sp)
        if self.stype=='normal' or self.stride==1:
            outs.append(spx[:, self.nums*self.width: (self.nums+1)*self.width])
        elif self.stype=='stage':
            outs.append(self.pool(spx[:, self.nums*self.width: (self.nums+1)*self.width]))
        out = concat(outs, 1)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Res2Net(Module):
    def __init__(self, block, layers, output_stride, baseWidth = 26, scale = 4):
        super(Res2Net, self).__init__()
        self.baseWidth = baseWidth
        self.scale = scale
        self.inplanes = 64
        blocks = [1, 2, 4]
        if output_stride == 16:
            strides = [1, 2, 2, 1]
            dilations = [1, 1, 1, 2]
        elif output_stride == 8:
            strides = [1, 2, 1, 1]
            dilations = [1, 1, 2, 4]
        else:
            raise NotImplementedError

        # Modules
        self.conv1 = nn.Sequential(
            nn.Conv(3, 32, 3, 2, 1, bias=False),
            nn.BatchNorm(32),
            nn.ReLU(),
            nn.Conv(32, 32, 3, 1, 1, bias=False),
            nn.BatchNorm(32),
            nn.ReLU(),
            nn.Conv(32, 64, 3, 1, 1, bias=False)
        )
        self.bn1 = nn.BatchNorm(64)
        self.relu = nn.ReLU()
        # self.maxpool = nn.Pool(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(block, 64, layers[0], stride=strides[0], dilation=dilations[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=strides[1], dilation=dilations[1])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=strides[2], dilation=dilations[2])
        self.layer4 = self._make_MG_unit(block, 512, blocks=blocks, stride=strides[3], dilation=dilations[3])


    def _make_layer(self, block, planes, blocks, stride=1, dilation=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Pool(kernel_size=stride, stride=stride, 
                    ceil_mode=True, op='mean'),
                nn.Conv(self.inplanes, planes * block.expansion, 
                    kernel_size=1, stride=1, bias=False),
                nn.BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, dilation, downsample,
                        stype='stage', baseWidth = self.baseWidth, scale=self.scale))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation, baseWidth = self.baseWidth, scale=self.scale))

        return nn.Sequential(*layers)

    def _make_MG_unit(self, block, planes, blocks, stride=1, dilation=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Pool(kernel_size=stride, stride=stride, 
                    ceil_mode=True, op='mean'),
                nn.Conv(self.inplanes, planes * block.expansion, 
                    kernel_size=1, stride=1, bias=False),
                nn.BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, dilation=blocks[0]*dilation,
                            downsample=downsample, stype='stage', baseWidth = self.baseWidth, scale=self.scale))
        self.inplanes = planes * block.expansion
        for i in range(1, len(blocks)):
            layers.append(block(self.inplanes, planes, stride=1,
                                dilation=blocks[i]*dilation, baseWidth = self.baseWidth, scale=self.scale))

        return nn.Sequential(*layers)

    def execute(self, input):

        x = self.conv1(input)
        x = self.bn1(x)

        x = self.relu(x)
        x = argmax_pool(x, 2, 2)
        x = self.layer1(x)
        low_level_feat = x
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.layer4(x)
        return x, low_level_feat
    
def res2net50(output_stride):
    model = Res2Net(Bottle2neck, [3,4,6,3], output_stride)
    return model

def res2net101(output_stride):
    model = Res2Net(Bottle2neck, [3,4,23,3], output_stride)
    return model



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import multiprocessing as mp
import numpy as np
import ctypes
import random
import pickle
import ctypes

recv_raw_call = 0.0

class RingBufferAllocator:
    def __init__(self, size):
        self.size = size
        self.l = mp.Value(ctypes.c_longlong, 0, lock=False)
        self.r = mp.Value(ctypes.c_longlong, 0, lock=False)
        self.is_full = mp.Value(ctypes.c_bool, False, lock=False)
        self.lock = mp.Lock()
        self.cv = mp.Condition(self.lock)
        
    def __repr__(self):
        l = self.l.value
        r = self.r.value
        is_full = self.is_full.value
        if is_full:
            cap = 0
        else:
            cap = (r - l) / self.size
            if cap<=0: cap += 1
        return f"Buffer(free={cap*100:.3f}% l={l} r={r} size={self.size})"
    
    def alloc_with_lock(self, size):
        with self.lock:
            while True:
                location = self.alloc(size)
                if location is not None: break
                self.cv.wait()
        return location
    
    def free_with_lock(self, size):
        with self.lock:
            location = self.free(size)
            self.cv.notify()
            return location

    def clear(self):
        with self.lock:
            self.l.value = 0
            self.r.value = 0
            self.is_full.value = False
    
    def alloc(self, size):
        if size > self.size:
            raise RuntimeError(f"Buffer size too small {self.size}<{size}")
        l = self.l.value
        r = self.r.value
        is_full = self.is_full.value
        if is_full: return None
        if l == r and l > 0:
            self.l.value = self.r.value = l = r = 0
        # [l, r)
        if r > l:
            freed = r - l
            if freed < size:
                # |----l......r---|
                # |----#########--|
                return None
            # |----l......r---|
            # |----#####------|
            location = l
            self.l.value = l = l + size
        else:
            freed = self.size - l
            if freed < size:
                # |.....r------l...|
                # |------------#######
                if size > r:
                    # |.....r------l...|
                    # |#######-----------
                    return None
                # |.....r------l...|
                # |#####-----------
                if size == r:
                    self.is_full.value = is_full= True
                location = 0
                self.l.value = l = size
            else:
                # |.....r------l...|
                # |------------##--|
                location = l
                if freed == size:
                    self.l.value = l = 0
                else:
                    self.l.value = l = l + size
        if l == r:
            self.is_full.value = is_full = True
        return location
    
    def free(self, size):
        l = self.l.value
        r = self.r.value
        is_full = self.is_full.value
        if size==0: return r
        if is_full:
            self.is_full.value = is_full = False
        elif l == r:
            return None
        location = r
        self.r.value = r = r + size
        if r > self.size:
            location = 0
            self.r.value = r = size
        elif r == self.size:
            self.r.value = r = 0
        return location
                
def str_to_char_array(s, array_len):
    if len(s) > array_len: s = s[:array_len]
    a = np.array(s, dtype='c')
    if len(s) < array_len:
        a = np.pad(a, (0,array_len-len(s)), constant_values=' ')
    return a

def char_array_to_str(a):
    return str(a.tostring(), 'ascii').strip()

class RingBuffer:
    def __init__(self, buffer):
        self.allocator = RingBufferAllocator(len(buffer))
        self.buffer = buffer

    def clear(self): self.allocator.clear()
            
    def send_int(self, data):
        # int: int64[1]
        #      data
        self.send_raw(np.array([data], dtype='int64'))
    def recv_int(self):
        return int(self.recv_raw(8, (1,), 'int64')[0])
    
    def send_float(self, data):
        # float: float64[1]
        #         data
        self.send_raw(np.array([data], dtype='float64'))
    def recv_float(self):
        return float(self.recv_raw(8, (1,), 'float64')[0])
    
    def send_str(self, data):
        # str: int64[1] char[len]
        #        len     data
        data = np.array(data, dtype='c')
        self.send_int(data.nbytes)
        self.send_raw(data)
    def recv_str(self):
        nbytes = self.recv_int()
        data = self.recv_raw(nbytes, nbytes, 'c')
        return str(data.tostring(), 'ascii')
    
    def send_ndarray(self, data):
        # str: int64[1]  char[8]  int64[1]  int64[slen] char[nbytes]
        #        slen     dtype   nbytes     shape       data
        shape = data.shape
        slen = len(shape)
        self.send_int(slen)
        self.send_fix_len_str(str(data.dtype))
        self.send_int(data.nbytes)
        self.send_raw(np.array(shape, dtype='int64'))
        self.send_raw(data)

    def recv_ndarray(self):
        slen = self.recv_int()
        dtype = self.recv_fix_len_str()
        nbytes = self.recv_int()
        shape = self.recv_raw(slen*8, slen, 'int64')
        data = self.recv_raw(nbytes, shape, dtype)
        return data
    
    def send_tuple(self, data):
        # tuple: int64[1] ....
        #         len
        length = len(data)
        self.send_int(length)
        for a in data:
            self.send(a)
    def recv_tuple(self):
        length = self.recv_int()
        return tuple(self.recv() for i in range(length))
    
    def send_list(self, data):
        # list: int64[1] ....
        #         len
        length = len(data)
        self.send_int(length)
        for a in data:
            self.send(a)

    def recv_list(self):
        length = self.recv_int()
        return [self.recv() for i in range(length)]

    def send_pickle(self, data):
        # pickle: int64[1] char[len]
        #         len     data
        data = pickle.dumps(data)
        data = np.frombuffer(data, dtype='c')
        self.send_int(data.nbytes)
        self.send_raw(data)

    def recv_pickle(self):
        nbytes = self.recv_int()
        data = self.recv_raw(nbytes, nbytes, 'c')
        return pickle.loads(data.tostring())
        
    def __repr__(self):
        return f"{self.allocator}@0x{hex(ctypes.addressof(self.buffer))}"
    
    def send_raw(self, data):
        assert isinstance(data, np.ndarray) # and data.flags.c_contiguous
        with self.allocator.lock:
            location = self.allocator.alloc(data.nbytes)
            while location is None:
                self.allocator.cv.wait()
                location = self.allocator.alloc(data.nbytes)
            window = np.ndarray(shape=data.shape, dtype=data.dtype, 
                                buffer=self.buffer, offset=location)
            window[:] = data
            self.allocator.cv.notify()
            assert window.nbytes == data.nbytes
        
    def recv_raw(self, nbytes, shape, dtype):
        global recv_raw_call
        recv_raw_call += 1
        with self.allocator.lock:
            location = self.allocator.free(nbytes)
            while location is None:
                self.allocator.cv.wait()
                location = self.allocator.free(nbytes)
            data = np.ndarray(shape=shape, dtype=dtype, 
                              buffer=self.buffer, offset=location).copy()
            self.allocator.cv.notify()
            assert data.nbytes == nbytes
            return data
    
    def send_fix_len_str(self, s, array_len=8):
        data = str_to_char_array(s, array_len)
        self.send_raw(data)
    
    def recv_fix_len_str(self, array_len=8):
        data = self.recv_raw(8, 8, 'c')
        return char_array_to_str(data)
    
    def send(self, data):
        ts = type(data).__name__
        send = getattr(self, "send_"+ts, self.send_pickle)
        self.send_fix_len_str(ts)
        send(data)
    
    def recv(self):
        ts = self.recv_fix_len_str()
        recv = getattr(self, "recv_"+ts, self.recv_pickle)
        return recv()    
    



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np

class TestWhereOp(unittest.TestCase):
    def test(self):
        assert (jt.where([0,1,0,1])[0].data == [1,3]).all()
        a, = jt.where([0,1,0,1])
        assert a.uncertain_shape==[-4]
        a.data
        assert a.uncertain_shape==[2]
        a,b = jt.where([[0,0,1],[1,0,0]])
        assert (a.data==[0,1]).all() and (b.data==[2,0]).all()

    def test_reindex_dep(self):
        a = jt.random([10])
        b, = (a>1).where()
        assert len(b.data)==0
        b, = (a>0.5).where()
        assert (b.data==np.where(a.data>0.5)).all()
        b = a.reindex_var((a>0.5).where())
        assert (b.data==a.data[a.data>0.5]).all()

    def test_binary_dep(self):
        a = jt.random([10])
        b, = (a>0.5).where()
        b = b+1
        assert (b.data==np.where(a.data>0.5)[0]+1).all()
        b, = (a>1).where()
        b = b+1
        assert (b.data==np.where(a.data>1)[0]+1).all()

    def test_self_dep(self):
        a = jt.random([100])
        x = a.reindex_var((a>0.1).where())
        x = x.reindex_var((x<0.9).where())
        na = a.data
        assert (na[np.logical_and(na>0.1, na<0.9)]==x.data).all()

    def test_reduce_dep(self):
        a = jt.random([100,100])
        index = (a>0.5).where()
        x = a.reindex_var(index)
        xsum =x.sum()
        na = a.data
        assert np.allclose(np.sum(na[na>0.5]),xsum.data), (x.data, xsum.data, np.sum(na[na>0.5]))
        
    def test_doc(self):
        assert "Where Operator" in jt.where.__doc__

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#    Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from jittor import nn

class TestOptimizer(unittest.TestCase):
    def test_param_groups(self):
        pa = jt.ones((1,))
        pb = jt.ones((1,))
        data = jt.ones((1,))
        opt = nn.SGD([
            {"params":[pa], "lr":0.1}, 
            {"params":[pb]}, 
        ], 1)
        opt.step(pa*data+pb*data)
        assert pa.data == 0.9 and pb.data == 0, (pa, pb)


if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import os
import re
import jittor as jt
from jittor import LOG

def find_log_with_re(logs, pattern=None, **args):
    if pattern:
        pattern = re.compile(pattern)
    flogs = []
    for log in logs:
        for arg in args:
            if log[arg] != args[arg]:
                break
        else:
            if pattern:
                res = re.findall(pattern, log["msg"])
                if len(res):
                    flogs.append(res[0])
            else:
                flogs.append(log["msg"])
    return flogs

class TestLog(unittest.TestCase):
    def test_log_capture(self):
        LOG.log_capture_start()
        with jt.flag_scope(log_v=1000, log_vprefix=""):
            LOG.v("1")
            LOG.vv("2")
            LOG.i("3")
            LOG.w("4")
            LOG.e("5")
            a = jt.zeros([10])
            a.sync()
        LOG.log_capture_stop()
        # TODO: why need manually delete this variable?
        del a
        logs = LOG.log_capture_read()
        logs2 = LOG.log_capture_read()
        assert len(logs2)==0

        for i in range(5):
            assert logs[i]['msg'] == str(i+1)
            assert logs[i]['level'] == 'iiiwe'[i]
            assert logs[i]['name'] == 'test_log.py'
        finished_log = [ l["msg"] for l in logs 
            if l["name"]=="executor.cc" and "return vars:" in l["msg"]]
        assert len(finished_log)==1 and "[10,]" in finished_log[0]


if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors:
#     Guowei Yang <471184555@qq.com>
#     Wenyang Zhou <576825820@qq.com>
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>.
#
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
from jittor import init, Module
import numpy as np
import math

class Pool(Module):
    def __init__(self, kernel_size, stride=None, padding=0, dilation=None, return_indices=None, ceil_mode=False, count_include_pad=True, op="maximum"):
        assert dilation == None
        assert return_indices == None
        self.kernel_size = kernel_size
        self.op = op
        self.stride = stride if stride else kernel_size
        self.padding = padding
        self.ceil_mode = ceil_mode
        self.count_include_pad = count_include_pad and padding != 0

    def execute(self, x):
        N,C,H,W = x.shape
        if self.ceil_mode == False:
            h = (H+self.padding*2-self.kernel_size)//self.stride+1
            w = (W+self.padding*2-self.kernel_size)//self.stride+1
        else:
            h = (H+self.padding*2-self.kernel_size + self.stride - 1)//self.stride+1
            w = (W+self.padding*2-self.kernel_size + self.stride - 1)//self.stride+1

        if self.op in ['maximum', 'minimum', 'mean']:
            if self.op == 'mean':
                if self.count_include_pad:
                    count = f"int count = {self.kernel_size*self.kernel_size};"
                else:
                    count = "int count = (k2_ - k2) * (k3_ - k3);"
            else:
                count = ""
            forward_body = f'''{{
                int k3 = i3*{self.stride}-{self.padding};
                int k2 = i2*{self.stride}-{self.padding};
                int k3_ = min(k3 + {self.kernel_size}, in0_shape3);
                int k2_ = min(k2 + {self.kernel_size}, in0_shape2);
                k3 = max(0, k3);
                k2 = max(0, k2);
                @out(i0, i1, i2, i3) = init_{self.op}(out_type);
                {count}
                for (int p = k2; p < k2_; ++p)
                    for (int q = k3; q < k3_; ++q)
                        @out(i0, i1, i2, i3) = {self.op}(out_type, @out(i0, i1, i2, i3), @in0(i0, i1, p, q));
            }}'''
            backward_body = f'''{{
                int k3 = i3*{self.stride}-{self.padding};
                int k2 = i2*{self.stride}-{self.padding};
                int k3_ = min(k3 + {self.kernel_size}, in0_shape3);
                int k2_ = min(k2 + {self.kernel_size}, in0_shape2);
                k3 = max(0, k3);
                k2 = max(0, k2);
                {count}
                int bo=1;
                for (int p = k2; p < k2_ && bo; ++p)
                    for (int q = k3; q < k3_ && bo; ++q) {{
                        {"atomicAdd(&@out(i0,i1,p,q), @dout(i0,i1,i2,i3)/count);"
                            if self.op == "mean" else
                        f"""if (@pout(i0,i1,i2,i3) == @in0(i0,i1,p,q)) {{
                            atomicAdd(&@out(i0,i1,p,q), @dout(i0,i1,i2,i3)),
                            bo=0;
                        }}"""}
                    }}
            }}'''
            out = jt.code([N,C,h,w], x.dtype, [x],
                cuda_header="""
                    #include <ops/binary_op_defs.h>
                    #include <misc/cuda_limits.h>
                """,
                cuda_src=f'''
                    __global__ static void kernel1(@ARGS_DEF) {{
                        @PRECALC
                        int p3 = threadIdx.x;
                        int s3 = blockDim.x;
                        int p2 = threadIdx.y + blockIdx.x * blockDim.y;
                        int s2 = blockDim.y * gridDim.x;
                        int i1 = blockIdx.y;
                        int i0 = blockIdx.z;
                        for (int i3 = p3; i3 < out_shape3; i3 += s3)
                        for (int i2 = p2; i2 < out_shape2; i2 += s2)
                            {forward_body}
                    }}
                    int tx = min(1024, out_shape3);
                    int ty = min(1024 / tx, out_shape2);
                    int bx = (out_shape2 - 1) / ty + 1;
                    int by = out_shape1;
                    int bz = out_shape0;
                    dim3 s1(bx, by, bz);
                    dim3 s2(tx, ty);
                    kernel1<<<s1, s2>>>(@ARGS);
                ''',
                cuda_grad_src=[f'''
                    __global__ static void kernel3(@ARGS_DEF) {{
                        @PRECALC
                        int p3 = threadIdx.x;
                        int s3 = blockDim.x;
                        int p2 = threadIdx.y + blockIdx.x * blockDim.y;
                        int s2 = blockDim.y * gridDim.x;
                        int i1 = blockIdx.y;
                        int i0 = blockIdx.z;
                        for (int i3 = p3; i3 < pout_shape3; i3 += s3)
                            for (int i2 = p2; i2 < pout_shape2; i2 += s2)
                                {backward_body}
                    }}
                    cudaMemsetAsync(out_p, 0, out->size);
                    int tx = min(1024, pout_shape3);
                    int ty = min(1024 / tx, pout_shape2);
                    int bx = (pout_shape2 - 1) / ty + 1;
                    int by = pout_shape1;
                    int bz = pout_shape0;
                    dim3 s1_(bx, by, bz);
                    dim3 s2_(tx, ty);
                    kernel3<<<s1_, s2_>>>(@ARGS);
                '''],
                cpu_header='#include <ops/binary_op_defs.h>',
                cpu_src=f'''
                    using namespace std;
                    for (int i0=0; i0<out_shape0; i0++)
                    for (int i1=0; i1<out_shape1; i1++)
                    for (int i2=0; i2<out_shape2; i2++)
                    for (int i3=0; i3<out_shape3; i3++)
                        {forward_body}
                ''',
                cpu_grad_src = [f'''
                    using namespace std;
                    std::memset(out_p, 0, out->size);
                    #define atomicAdd(a,b) (*a) += b

                    for (int i0=0; i0<pout_shape0; i0++)
                    for (int i1=0; i1<pout_shape1; i1++)
                    for (int i2=0; i2<pout_shape2; i2++) 
                    for (int i3=0; i3<pout_shape3; i3++)
                        {backward_body}
                '''])
            return out
        else:
            # TODO: backward 
            xx = x.reindex([N,C,h,w,self.kernel_size,self.kernel_size], [
                "i0", # Nid
                "i1", # Cid
                f"i2*{self.stride}-{self.padding}+i4", # Hid
                f"i3*{self.stride}-{self.padding}+i5", # Wid
            ])
            return xx.reduce(self.op, [4,5])


class AdaptiveAvgPool2d(Module):
    def __init__(self, output_size):
        self.output_size = output_size

    def execute(self, x):
        if isinstance(self.output_size, int):
            oh = self.output_size
            ow = self.output_size
        elif isinstance(self.output_size, tuple) or isinstance(self.output_size, list):
            oh = x.shape[2] if self.output_size[0] is None else self.output_size[0]
            ow = x.shape[3] if self.output_size[1] is None else self.output_size[1]
        else:
            raise TypeError(f"AdaptiveAvgPool2d only support int, typle or list input. Not support {type(self.output_size)} yet.")
        N,C,H,W = x.shape
        self.sh = math.floor(H / oh)
        self.sw = math.floor(W / ow)
        self.ksh = H - (oh - 1) * self.sh
        self.ksw = W - (ow - 1) * self.sw
        h = (H-self.ksh)//self.sh+1
        w = (W-self.ksw)//self.sw+1
        xx = x.reindex([N,C,h,w,self.ksh,self.ksw], [
            "i0", # Nid
            "i1", # Cid
            f"i2*{self.sh}+i4", # Hid
            f"i3*{self.sw}+i5", # Wid
        ])
        return xx.reduce("mean", [4,5])

def pool(x, kernel_size, op, padding=0, stride = 1):
    return Pool(kernel_size, stride, padding, op=op)(x)